"""
awslimitchecker/tests/test_trustedadvisor.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

##############################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
##############################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
##############################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
##############################################################################
"""

import sys
from botocore.exceptions import ClientError
from awslimitchecker.trustedadvisor import TrustedAdvisor, datetime_now
from awslimitchecker.services.base import _AwsService
from awslimitchecker.limit import AwsLimit
import pytest
from datetime import datetime
from freezegun import freeze_time
from pytz import utc

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock
else:
    from unittest.mock import patch, call, Mock


pbm = 'awslimitchecker.trustedadvisor'
pb = '%s.TrustedAdvisor' % pbm


class TestInit(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        cls = TrustedAdvisor({}, {})
        assert cls.conn is None
        assert cls._boto3_connection_kwargs == {
            'region_name': 'us-east-1'
        }
        assert cls.all_services == {}
        assert cls.limits_updated is False
        assert cls.refresh_mode is None
        assert cls.refresh_timeout is None

    def test_boto_kwargs(self):
        mock_svc = Mock(spec_set=_AwsService)
        mock_svc.get_limits.return_value = {}
        boto_args = dict(region_name='myregion',
                         aws_access_key_id='myaccesskey',
                         aws_secret_access_key='mysecretkey',
                         aws_session_token='mytoken')

        cls = TrustedAdvisor(
            {'foo': mock_svc},
            boto_args,
            ta_refresh_mode=123,
            ta_refresh_timeout=456
        )
        assert cls.conn is None
        cls_boto_args = cls._boto3_connection_kwargs
        assert cls_boto_args.get('region_name') == 'us-east-1'
        assert cls_boto_args.get('aws_access_key_id') == 'myaccesskey'
        assert cls_boto_args.get('aws_secret_access_key') == 'mysecretkey'
        assert cls_boto_args.get('aws_session_token') == 'mytoken'
        assert cls.ta_region == 'myregion'
        assert cls.all_services == {'foo': mock_svc}
        assert cls.limits_updated is False
        assert cls.refresh_mode == 123
        assert cls.refresh_timeout == 456


class TestUpdateLimits(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        mock_results = Mock()
        with patch('%s.connect' % pb, autospec=True) as mock_connect:
            with patch('%s._poll' % pb, autospec=True) as mock_poll:
                with patch('%s._update_services' % pb,
                           autospec=True) as mock_update_services:
                    mock_poll.return_value = mock_results
                    self.cls.update_limits()
        assert mock_connect.mock_calls == [call(self.cls)]
        assert mock_poll.mock_calls == [call(self.cls)]
        assert mock_update_services.mock_calls == [
            call(self.cls, mock_results)
        ]

    def test_again(self):
        mock_results = Mock()
        self.cls.limits_updated = True
        with patch('%s.connect' % pb, autospec=True) as mock_connect:
            with patch('%s._poll' % pb, autospec=True) as mock_poll:
                with patch('%s._update_services' % pb,
                           autospec=True) as mock_update_services:
                    with patch('%s.logger' % pbm) as mock_logger:
                        mock_poll.return_value = mock_results
                        self.cls.update_limits()
        assert mock_connect.mock_calls == []
        assert mock_poll.mock_calls == []
        assert mock_update_services.mock_calls == []
        assert mock_logger.mock_calls == [
            call.debug('Already polled TA; skipping update')
        ]


class TestGetLimitCheckId(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        api_resp = {
            'checks': [
                {
                    'category': 'performance',
                    'name': 'Service Limits',
                    'id': 'bar',
                    'metadata': [
                        'Region',
                        'Service',
                        'Limit Name',
                        'Limit Amount',
                        'Current Usage',
                        'Status'
                    ],
                },
                {
                    'category': 'fault_tolerance',
                    'name': 'ELB Cross-Zone Load Balancing',
                    'id': 'foo',
                },
            ]
        }
        self.mock_conn.describe_trusted_advisor_checks.return_value = api_resp
        res = self.cls._get_limit_check_id()
        assert res == (
            'bar',
            [
                'Region',
                'Service',
                'Limit Name',
                'Limit Amount',
                'Current Usage',
                'Status'
            ]
        )
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_checks(language='en')
        ]

    def test_none(self):
        api_resp = {
            'checks': [
                {
                    'category': 'performance',
                    'name': 'Something Else',
                    'id': 'bar',
                },
                {
                    'category': 'fault_tolerance',
                    'name': 'ELB Cross-Zone Load Balancing',
                    'id': 'foo',
                },
            ]
        }
        self.mock_conn.describe_trusted_advisor_checks.return_value = api_resp
        res = self.cls._get_limit_check_id()
        assert res == (None, None)
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_checks(language='en')
        ]

    def test_subscription_required(self):

        def se_api(language=None):
            response = {
                'ResponseMetadata': {
                    'HTTPStatusCode': 400,
                    'RequestId': '3cc9b2a8-c6e5-11e5-bc1d-b13dcea36176'
                },
                'Error': {
                    'Message': 'AWS Premium Support Subscription is required '
                               'to use this service.',
                    'Code': 'SubscriptionRequiredException'
                }
            }
            raise ClientError(response, 'operation')

        assert self.cls.have_ta is True
        self.mock_conn.describe_trusted_advisor_checks.side_effect = se_api
        with patch('awslimitchecker.trustedadvisor'
                   '.logger', autospec=True) as mock_logger:
            res = self.cls._get_limit_check_id()
        assert self.cls.have_ta is False
        assert res == (None, None)
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_checks(language='en')
        ]
        assert mock_logger.mock_calls == [
            call.debug("Querying Trusted Advisor checks"),
            call.warning("Cannot check TrustedAdvisor: %s",
                         'AWS Premium Support Subscription is required to '
                         'use this service.')
        ]

    def test_other_exception(self):

        def se_api(language=None):
            response = {
                'ResponseMetadata': {
                    'HTTPStatusCode': 400,
                    'RequestId': '3cc9b2a8-c6e5-11e5-bc1d-b13dcea36176'
                },
                'Error': {
                    'Message': 'foo',
                    'Code': 'SomeOtherException'
                }
            }
            raise ClientError(response, 'operation')

        self.mock_conn.describe_trusted_advisor_checks.side_effect = se_api
        with pytest.raises(ClientError) as excinfo:
            self.cls._get_limit_check_id()
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_checks(language='en')
        ]
        assert excinfo.value.response['ResponseMetadata'][
                   'HTTPStatusCode'] == 400
        assert excinfo.value.response['Error']['Message'] == 'foo'
        assert excinfo.value.response['Error']['Code'] == 'SomeOtherException'


class TestPoll(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_none(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_result
        with patch('%s._get_limit_check_id' % pb, autospec=True) as mock_id:
            mock_id.return_value = None
            res = self.cls._poll()
        assert tmp.mock_calls == []
        assert res is None

    def test_basic(self):
        poll_return_val = {
            'result': {
                'timestamp': '2015-06-15T20:27:42Z',
                'flaggedResources': [
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-west-2',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '20',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'EC2',
                            'On-Demand instances - t2.micro',
                            'Unlimited',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'EC2',
                            'On-Demand instances - t2.small',
                            'error',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid2',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'AutoScaling',
                            'Launch configurations',
                            '20',
                            '18',
                            'Yellow'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid3',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '40',
                            '10',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid4',
                        'isSuppressed': False,
                        'metadata': [
                            '-',
                            'IAM',
                            'Users',
                            '5000',
                            '152',
                            'Green'
                        ]
                    },
                ]
            }
        }
        with patch('%s._get_limit_check_id' % pb, autospec=True) as mock_id:
            with patch('%s._get_refreshed_check_result' % pb,
                       autospec=True) as mock_hr:
                mock_hr.return_value = poll_return_val
                mock_id.return_value = (
                    'foo',
                    [
                        'Region',
                        'Service',
                        'Limit Name',
                        'Limit Amount',
                        'Current Usage',
                        'Status'
                    ]
                )
                res = self.cls._poll()
        assert self.mock_conn.mock_calls == []
        assert mock_id.mock_calls == [call(self.cls)]
        assert mock_hr.mock_calls == [
            call(self.cls, 'foo')
        ]
        assert res == {
            'AutoScaling': {
                'Launch configurations': 20,
                'Auto Scaling groups': 40,
            },
            'EC2': {
                'On-Demand instances - t2.micro': 'Unlimited'
            },
            'IAM': {
                'Users': 5000
            }
        }

    def test_no_timestamp(self):
        poll_return_val = {
            'result': {
                'flaggedResources': [
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-west-2',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '20',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'EC2',
                            'On-Demand instances - t2.micro',
                            'Unlimited',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'EC2',
                            'On-Demand instances - t2.small',
                            'error',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid2',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'AutoScaling',
                            'Launch configurations',
                            '20',
                            '18',
                            'Yellow'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid3',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '40',
                            '10',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid4',
                        'isSuppressed': False,
                        'metadata': [
                            '-',
                            'IAM',
                            'Users',
                            '5000',
                            '152',
                            'Green'
                        ]
                    },
                ]
            }
        }
        with patch('%s._get_limit_check_id' % pb, autospec=True) as mock_id:
            with patch('%s._get_refreshed_check_result' % pb,
                       autospec=True) as mock_hr:
                mock_hr.return_value = poll_return_val
                mock_id.return_value = (
                    'foo',
                    [
                        'Region',
                        'Service',
                        'Limit Name',
                        'Limit Amount',
                        'Current Usage',
                        'Status'
                    ]
                )
                res = self.cls._poll()
        assert self.mock_conn.mock_calls == []
        assert mock_id.mock_calls == [call(self.cls)]
        assert mock_hr.mock_calls == [
            call(self.cls, 'foo')
        ]
        assert res == {
            'AutoScaling': {
                'Launch configurations': 20,
                'Auto Scaling groups': 40,
            },
            'EC2': {
                'On-Demand instances - t2.micro': 'Unlimited'
            },
            'IAM': {
                'Users': 5000
            }
        }

    def test_region(self):
        self.cls.ta_region = 'us-west-2'
        poll_return_value = {
            'result': {
                'timestamp': '2015-06-15T20:27:42Z',
                'flaggedResources': [
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-west-2',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '20',
                            '2',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid2',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-east-1',
                            'AutoScaling',
                            'Launch configurations',
                            '20',
                            '18',
                            'Yellow'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid3',
                        'isSuppressed': False,
                        'region': 'us-east-1',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '40',
                            '10',
                            'Green'
                        ]
                    },
                    {
                        'status': 'ok',
                        'resourceId': 'resid4',
                        'isSuppressed': False,
                        'metadata': [
                            '-',
                            'IAM',
                            'Users',
                            '5000',
                            '152',
                            'Green'
                        ]
                    },
                ]
            }
        }
        with patch('%s._get_limit_check_id' % pb, autospec=True) as mock_id:
            with patch('%s._get_refreshed_check_result' % pb,
                       autospec=True) as mock_hr:
                mock_hr.return_value = poll_return_value
                mock_id.return_value = (
                    'foo',
                    [
                        'Region',
                        'Service',
                        'Limit Name',
                        'Limit Amount',
                        'Current Usage',
                        'Status'
                    ]
                )
                res = self.cls._poll()
        assert self.mock_conn.mock_calls == []
        assert mock_id.mock_calls == [call(self.cls)]
        assert mock_hr.mock_calls == [
            call(self.cls, 'foo')
        ]
        assert res == {
            'AutoScaling': {
                'Auto Scaling groups': 20,
            },
            'IAM': {
                'Users': 5000
            }
        }

    def test_dont_have_ta(self):
        self.cls.have_ta = False
        with patch('%s._get_limit_check_id' % pb, autospec=True) as mock_id:
            with patch('%s._get_refreshed_check_result' % pb,
                       autospec=True) as mock_hr:
                with patch('awslimitchecker.trustedadvisor'
                           '.logger', autospec=True) as mock_logger:
                    res = self.cls._poll()
        assert self.mock_conn.mock_calls == []
        assert mock_id.mock_calls == [
            call(self.cls)
        ]
        assert mock_hr.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info('Beginning TrustedAdvisor poll'),
            call.info('TrustedAdvisor.have_ta is False; not polling TA')
        ]
        assert res == {}


class TestGetRefreshedCheckResult(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_mode_none(self):
        self.cls.refresh_mode = None
        with patch('%s._get_check_result' % pb, autospec=True) as mock_gcr:
            with patch('%s._can_refresh_check' % pb, autospec=True) as mock_crc:
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s._poll_for_refresh' % pb,
                               autospec=True) as mock_pfr:
                        mock_gcr.return_value = ({'mock': 'gcr'}, None)
                        res = self.cls._get_refreshed_check_result('abc123')
        assert res == {'mock': 'gcr'}
        assert mock_gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_crc.mock_calls == []
        assert mock_pfr.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info("Not refreshing Trusted Advisor check (refresh mode is "
                      "None)")
        ]

    @freeze_time("2016-12-16 10:40:42", tz_offset=0)
    def test_mode_int(self):
        self.cls.refresh_mode = 120  # 2 minutes
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        with patch('%s._get_check_result' % pb, autospec=True) as mock_gcr:
            with patch('%s._can_refresh_check' % pb, autospec=True) as mock_crc:
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s._poll_for_refresh' % pb,
                               autospec=True) as mock_pfr:
                        mock_gcr.return_value = ({'mock': 'gcr'}, check_dt)
                        mock_pfr.return_value = {'mock': 'pfr'}
                        mock_crc.return_value = True
                        res = self.cls._get_refreshed_check_result('abc123')
        assert res == {'mock': 'pfr'}
        assert mock_gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_crc.mock_calls == [call(self.cls, 'abc123')]
        assert mock_pfr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_logger.mock_calls == [
            call.debug('Handling refresh of check: %s', 'abc123'),
            call.debug('ta_refresh_mode older; check last refresh: %s; '
                       'threshold=%d seconds', check_dt, 120),
            call.info('Refreshing Trusted Advisor check: %s', 'abc123')
        ]

    @freeze_time("2016-12-16 10:40:42", tz_offset=0)
    def test_mode_int_within_threshold(self):
        self.cls.refresh_mode = 120  # 2 minutes
        check_dt = datetime(2016, 12, 16, hour=10, minute=40, second=12,
                            tzinfo=utc)
        with patch('%s._get_check_result' % pb, autospec=True) as mock_gcr:
            with patch('%s._can_refresh_check' % pb, autospec=True) as mock_crc:
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s._poll_for_refresh' % pb,
                               autospec=True) as mock_pfr:
                        mock_gcr.return_value = ({'mock': 'gcr'}, check_dt)
                        mock_pfr.return_value = {'mock': 'pfr'}
                        mock_crc.return_value = True
                        res = self.cls._get_refreshed_check_result('abc123')
        assert res == {'mock': 'gcr'}
        assert mock_gcr.mock_calls == [
            call(self.cls, 'abc123'),
            call(self.cls, 'abc123')
        ]
        assert mock_crc.mock_calls == [call(self.cls, 'abc123')]
        assert mock_pfr.mock_calls == []
        assert mock_logger.mock_calls == [
            call.debug('Handling refresh of check: %s', 'abc123'),
            call.debug('ta_refresh_mode older; check last refresh: %s; '
                       'threshold=%d seconds', check_dt, 120),
            call.warning('Trusted Advisor check %s last refresh time of %s '
                         'is newer than refresh threshold of %d seconds.',
                         'abc123',
                         datetime(2016, 12, 16, 10, 40, 12, tzinfo=utc),
                         120)
        ]

    @freeze_time("2016-12-16 10:40:42", tz_offset=0)
    def test_mode_trigger(self):
        self.cls.refresh_mode = 'trigger'
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        with patch('%s._get_check_result' % pb, autospec=True) as mock_gcr:
            with patch('%s._can_refresh_check' % pb, autospec=True) as mock_crc:
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s._poll_for_refresh' % pb,
                               autospec=True) as mock_pfr:
                        mock_gcr.return_value = ({'mock': 'gcr'}, check_dt)
                        mock_pfr.return_value = {'mock': 'pfr'}
                        mock_crc.return_value = True
                        res = self.cls._get_refreshed_check_result('abc123')
        assert res == {'mock': 'gcr'}
        assert mock_gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_crc.mock_calls == [call(self.cls, 'abc123')]
        assert mock_pfr.mock_calls == []
        assert mock_logger.mock_calls == [
            call.debug('Handling refresh of check: %s', 'abc123'),
            call.info('Refreshing Trusted Advisor check: %s', 'abc123')
        ]

    @freeze_time("2016-12-16 10:40:42", tz_offset=0)
    def test_cant_refresh(self):
        self.cls.refresh_mode = 120  # 2 minutes
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        with patch('%s._get_check_result' % pb, autospec=True) as mock_gcr:
            with patch('%s._can_refresh_check' % pb, autospec=True) as mock_crc:
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s._poll_for_refresh' % pb,
                               autospec=True) as mock_pfr:
                        mock_gcr.return_value = ({'mock': 'gcr'}, check_dt)
                        mock_pfr.return_value = {'mock': 'pfr'}
                        mock_crc.return_value = False
                        res = self.cls._get_refreshed_check_result('abc123')
        assert res == {'mock': 'gcr'}
        assert mock_gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_crc.mock_calls == [call(self.cls, 'abc123')]
        assert mock_pfr.mock_calls == []
        assert mock_logger.mock_calls == [
            call.debug('Handling refresh of check: %s', 'abc123')
        ]


class TestGetCheckResult(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_result
        check_result = {
            'result': {
                'timestamp': '2015-06-15T20:27:42Z',
                'flaggedResources': [
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-west-2',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '20',
                            '2',
                            'Green'
                        ]
                    }
                ]
            }
        }
        tmp.return_value = check_result
        res = self.cls._get_check_result('abc123')
        assert tmp.mock_calls == [
            call(checkId='abc123', language='en')
        ]
        assert res == (
            check_result,
            datetime(2015, 6, 15, 20, 27, 42, tzinfo=utc)
        )

    def test_no_timestamp(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_result
        check_result = {
            'result': {
                'flaggedResources': [
                    {
                        'status': 'ok',
                        'resourceId': 'resid1',
                        'isSuppressed': False,
                        'region': 'us-west-2',
                        'metadata': [
                            'us-west-2',
                            'AutoScaling',
                            'Auto Scaling groups',
                            '20',
                            '2',
                            'Green'
                        ]
                    }
                ]
            }
        }
        tmp.return_value = check_result
        res = self.cls._get_check_result('abc123')
        assert tmp.mock_calls == [
            call(checkId='abc123', language='en')
        ]
        assert res == (check_result, None)


class TestCanRefreshCheck(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

    def test_true(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        chkstat = {
            'checkId': 'abc123',
            'status': 'none',
            'millisUntilNextRefreshable': 0
        }
        tmp.return_value = {
            'statuses': [chkstat]
        }
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            res = self.cls._can_refresh_check('abc123')
        assert res is True
        assert tmp.mock_calls == [
            call(checkIds=['abc123'])
        ]
        assert mock_logger.mock_calls == [
            call.debug('TA Check %s refresh status: %s',
                       'abc123', chkstat)
        ]

    def test_false(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        chkstat = {
            'checkId': 'abc123',
            'status': 'none',
            'millisUntilNextRefreshable': 123456
        }
        tmp.return_value = {
            'statuses': [chkstat]
        }
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            res = self.cls._can_refresh_check('abc123')
        assert res is False
        assert tmp.mock_calls == [
            call(checkIds=['abc123'])
        ]
        assert mock_logger.mock_calls == [
            call.debug('TA Check %s refresh status: %s',
                       'abc123', chkstat),
            call.warning("Trusted Advisor check cannot be refreshed for "
                         "another %d milliseconds; skipping refresh and "
                         "getting check results now", 123456)
        ]

    def test_exception(self):
        tmp = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        chkstat = {
            'checkId': 'abc123',
            'status': 'none'
        }
        tmp.return_value = {
            'statuses': [chkstat]
        }
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            res = self.cls._can_refresh_check('abc123')
        assert res is True
        assert tmp.mock_calls == [
            call(checkIds=['abc123'])
        ]
        assert mock_logger.mock_calls == [
            call.debug('TA Check %s refresh status: %s',
                       'abc123', chkstat),
            call.warning('Could not get refresh status for TA check %s',
                         'abc123', exc_info=True)
        ]


class TestPollForRefresh(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

    def test_no_timeout(self):
        self.cls.refresh_timeout = None
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        now_dt = datetime(2016, 12, 16, hour=11, minute=30, second=12,
                          tzinfo=utc)
        statuses = [
            {'statuses': [{'status': 'none'}]},
            {'statuses': [{'status': 'enqueued'}]},
            {'statuses': [{'status': 'processing'}]},
            {'statuses': [{'status': 'success'}]}
        ]
        m_s = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.sleep' % pbm, autospec=True) as mock_sleep:
                with patch('%s._get_check_result' % pb, autospec=True) as gcr:
                    with patch('%s.datetime_now' % pbm) as mock_dt_now:
                        mock_dt_now.return_value = now_dt
                        m_s.side_effect = statuses
                        gcr.return_value = ({'foo': 'bar'}, check_dt)
                        res = self.cls._poll_for_refresh('abc123')
        assert res == {'foo': 'bar'}
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123'])
        ]
        assert gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_sleep.mock_calls == [
            call(30), call(30), call(30)
        ]
        assert mock_dt_now.mock_calls == [
            call(), call(), call(), call(), call()
        ]
        assert mock_logger.mock_calls == [
            call.warning('Polling for TA check %s refresh...', 'abc123'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'none'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'enqueued'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'processing'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; done polling', 'success'),
            call.info('Done polling for check refresh'),
            call.debug('Check shows last refresh time of: %s', check_dt)
        ]

    def test_timeout(self):
        self.cls.refresh_timeout = 45
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        now_dts = [
            datetime(2016, 12, 16, hour=11, minute=30, second=0, tzinfo=utc),
            datetime(2016, 12, 16, hour=11, minute=30, second=0, tzinfo=utc),
            datetime(2016, 12, 16, hour=11, minute=30, second=30, tzinfo=utc),
            datetime(2016, 12, 16, hour=11, minute=31, second=0, tzinfo=utc),
        ]
        status = {'statuses': [{'status': 'processing'}]}
        m_s = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.sleep' % pbm, autospec=True) as mock_sleep:
                with patch('%s._get_check_result' % pb, autospec=True) as gcr:
                    with patch('%s.datetime_now' % pbm) as mock_dt_now:
                        mock_dt_now.side_effect = now_dts
                        m_s.return_value = status
                        gcr.return_value = ({'foo': 'bar'}, check_dt)
                        res = self.cls._poll_for_refresh('abc123')
        assert res == {'foo': 'bar'}
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123'])
        ]
        assert gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_sleep.mock_calls == [
            call(30), call(30)
        ]
        assert mock_dt_now.mock_calls == [
            call(), call(), call(), call()
        ]
        assert mock_logger.mock_calls == [
            call.warning('Polling for TA check %s refresh...', 'abc123'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'processing'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'processing'),
            call.error('Timed out waiting for TA Check refresh; status=%s',
                       'processing'),
            call.info('Done polling for check refresh'),
            call.debug('Check shows last refresh time of: %s', check_dt)
        ]

    def test_none(self):
        self.cls.refresh_timeout = None
        check_dt = datetime(2016, 12, 16, hour=10, minute=30, second=12,
                            tzinfo=utc)
        now_dt = datetime(2016, 12, 16, hour=11, minute=30, second=12,
                          tzinfo=utc)
        statuses = [
            {'statuses': [{'status': 'none'}]},
            {'statuses': [{'status': 'enqueued'}]},
            {'statuses': [{'status': 'processing'}]},
            {'statuses': [{'status': 'none'}]}
        ]
        m_s = self.mock_conn.describe_trusted_advisor_check_refresh_statuses
        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.sleep' % pbm, autospec=True) as mock_sleep:
                with patch('%s._get_check_result' % pb, autospec=True) as gcr:
                    with patch('%s.datetime_now' % pbm) as mock_dt_now:
                        mock_dt_now.return_value = now_dt
                        m_s.side_effect = statuses
                        gcr.return_value = ({'foo': 'bar'}, check_dt)
                        res = self.cls._poll_for_refresh('abc123')
        assert res == {'foo': 'bar'}
        assert self.mock_conn.mock_calls == [
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123']),
            call.describe_trusted_advisor_check_refresh_statuses(
                checkIds=['abc123'])
        ]
        assert gcr.mock_calls == [call(self.cls, 'abc123')]
        assert mock_sleep.mock_calls == [
            call(30), call(30), call(30)
        ]
        assert mock_dt_now.mock_calls == [
            call(), call(), call(), call(), call()
        ]
        assert mock_logger.mock_calls == [
            call.warning('Polling for TA check %s refresh...', 'abc123'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'none'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'enqueued'),
            call.debug('Checking refresh status'),
            call.info('Refresh status: %s; sleeping 30s', 'processing'),
            call.debug('Checking refresh status'),
            call.warning('Trusted Advisor check refresh status went '
                         'from "%s" to "%s"; refresh is either complete '
                         'or timed out on AWS side. Continuing',
                         'processing', 'none'),
            call.info('Done polling for check refresh'),
            call.debug('Check shows last refresh time of: %s', check_dt)
        ]


class TestUpdateServices(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        mock_as_foo = Mock(spec_set=AwsLimit)
        mock_as_bar = Mock(spec_set=AwsLimit)
        mock_ec2_baz = Mock(spec_set=AwsLimit)
        mock_ec2_blarg = Mock(spec_set=AwsLimit)
        mock_vpc = Mock(spec_set=AwsLimit)
        ta_services = {
            'AutoScaling': {
                'foo': mock_as_foo,
                'bar': mock_as_bar
            },
            'EC2': {
                'baz': mock_ec2_baz,
                'blarg': mock_ec2_blarg
            },
            'VPC': {
                'VPC Elastic IP addresses (EIPs)': mock_vpc
            },
        }
        ta_results = {
            'AutoScaling': {
                'foo': 20,
                'bar': 40,
            },
            'EC2': {
                'baz': 5,
                'blam': 10,
                'blarg': 'Unlimited',
            },
            'OtherService': {
                'blarg': 1,
            },
            'VPC': {
                'VPC Elastic IP addresses (EIPs)': 11,
            }
        }
        with patch('awslimitchecker.trustedadvisor'
                   '.logger', autospec=True) as mock_logger:
            self.cls.ta_services = ta_services
            self.cls._update_services(ta_results)
        assert mock_logger.mock_calls == [
            call.debug("Updating TA limits on all services"),
            call.info("TrustedAdvisor returned check results for unknown "
                      "limit '%s' (service %s)", 'blam', 'EC2'),
            call.info("TrustedAdvisor returned check results for unknown "
                      "service '%s'", 'OtherService'),
            call.info("Done updating TA limits on all services"),
        ]
        assert mock_as_foo.mock_calls == [
            call._set_ta_limit(20)
        ]
        assert mock_as_bar.mock_calls == [
            call._set_ta_limit(40)
        ]
        assert mock_ec2_baz.mock_calls == [
            call._set_ta_limit(5)
        ]
        assert mock_ec2_blarg.mock_calls == [
            call._set_ta_unlimited()
        ]
        assert mock_vpc.mock_calls == [
            call._set_ta_limit(11)
        ]


class TestMakeTAServiceDict(object):

    def setup(self):
        self.mock_conn = Mock()
        self.mock_client_config = Mock()
        type(self.mock_client_config).region_name = 'us-east-1'
        type(self.mock_conn)._client_config = self.mock_client_config
        self.cls = TrustedAdvisor({}, {})
        self.cls.conn = self.mock_conn

        self.mock_svc1 = Mock(spec_set=_AwsService)
        self.mock_svc2 = Mock(spec_set=_AwsService)
        self.services = {
            'SvcFoo': self.mock_svc1,
            'SvcBar': self.mock_svc2,
        }

    def test_simple(self):
        mock_ec2 = Mock(spec_set=_AwsService)
        mock_el1 = Mock(spec_set=AwsLimit)
        type(mock_el1).name = 'el1'
        type(mock_el1).ta_service_name = 'EC2'
        type(mock_el1).ta_limit_name = 'el1'
        mock_el2 = Mock(spec_set=AwsLimit)
        type(mock_el2).name = 'el2'
        type(mock_el2).ta_service_name = 'Foo'
        type(mock_el2).ta_limit_name = 'el2'
        mock_ec2.get_limits.return_value = {
            'mock_el1': mock_el1,
            'mock_el2': mock_el2
        }

        mock_vpc = Mock(spec_set=_AwsService)
        mock_vl1 = Mock(spec_set=AwsLimit)
        type(mock_vl1).name = 'vl1'
        type(mock_vl1).ta_service_name = 'VPC'
        type(mock_vl1).ta_limit_name = 'other name'
        mock_vl2 = Mock(spec_set=AwsLimit)
        type(mock_vl2).name = 'vl2'
        type(mock_vl2).ta_service_name = 'Foo'
        type(mock_vl2).ta_limit_name = 'other limit'
        mock_vpc.get_limits.return_value = {
            'mock_vl1': mock_vl1,
            'mock_vl2': mock_vl2
        }

        svcs = {
            'EC2': mock_ec2,
            'VPC': mock_vpc
        }

        expected = {
            'EC2': {
                'el1': mock_el1
            },
            'VPC': {
                'other name': mock_vl1
            },
            'Foo': {
                'el2': mock_el2,
                'other limit': mock_vl2
            }
        }
        self.cls.all_services = svcs
        assert self.cls._make_ta_service_dict() == expected


class TestDatetimeNow(object):

    @freeze_time("2016-12-16 10:40:42")
    def test_it(self):
        dt = datetime(2016, 12, 16, hour=10, minute=40, second=42)
        assert datetime_now() == dt



"""
awslimitchecker/tests/services/test_ecs.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Di Zou <zou@pythian.com>
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from awslimitchecker.services.ecs import _EcsService

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, DEFAULT
else:
    from unittest.mock import patch, call, Mock, DEFAULT


pbm = 'awslimitchecker.services.ecs'  # module patch base
pb = '%s._EcsService' % pbm  # class patch pase


class Test_EcsService(object):

    def test_init(self):
        """test __init__()"""
        cls = _EcsService(21, 43)
        assert cls.service_name == 'ECS'
        assert cls.api_name == 'ecs'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _EcsService(21, 43)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Clusters',
            'Container Instances per Cluster',
            'EC2 Tasks per Service (desired count)',
            'Fargate Tasks',
            'Services per Cluster',
        ])
        for name, limit in res.items():
            assert limit.service == cls
            assert limit.def_warning_threshold == 21
            assert limit.def_critical_threshold == 43

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _EcsService(21, 43)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        with patch.multiple(
            pb,
            autospec=True,
            connect=DEFAULT,
            _find_usage_clusters=DEFAULT
        ) as mocks:
            cls = _EcsService(21, 43)
            assert cls._have_usage is False
            cls.find_usage()
        assert mocks['connect'].mock_calls == [call(cls)]
        assert cls._have_usage is True
        assert mocks['connect'].return_value.mock_calls == []

    def test_find_usage_clusters(self):
        def se_clusters(*_, **kwargs):
            if kwargs['clusters'] == ['c1arn']:
                return {
                    'clusters': [
                        {
                            'clusterArn': 'c1arn',
                            'clusterName': 'c1name',
                            'status': 'string',
                            'registeredContainerInstancesCount': 11,
                            'runningTasksCount': 6,
                            'pendingTasksCount': 45,
                            'activeServicesCount': 23,
                            'statistics': [
                                {'name': 'runningEC2TasksCount', 'value': '0'},
                                {
                                    'name': 'runningFargateTasksCount',
                                    'value': '4'
                                },
                                {'name': 'pendingEC2TasksCount', 'value': '0'},
                                {
                                    'name': 'pendingFargateTasksCount',
                                    'value': '2'
                                }
                            ]
                        }
                    ]
                }
            elif kwargs['clusters'] == ['c2arn']:
                return {
                    'clusters': [
                        {
                            'clusterArn': 'c2arn',
                            'clusterName': 'c2name',
                            'status': 'string',
                            'registeredContainerInstancesCount': 3,
                            'runningTasksCount': 8,
                            'pendingTasksCount': 22,
                            'activeServicesCount': 2
                        }
                    ]
                }
            return {}

        mock_conn = Mock()
        mock_conn.describe_clusters.side_effect = se_clusters
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = [{
            'clusterArns': [
                'c1arn',
                'c2arn'
            ],
            'nextToken': 'string'
        }]

        mock_conn.get_paginator.return_value = mock_paginator
        cls = _EcsService(21, 43)
        cls.conn = mock_conn
        with patch('%s._find_usage_one_cluster' % pb, autospec=True) as m_fuoc:
            cls._find_usage_clusters()
        assert mock_conn.mock_calls == [
            call.get_paginator('list_clusters'),
            call.get_paginator().paginate(),
            call.describe_clusters(
                clusters=['c1arn'], include=['STATISTICS']
            ),
            call.describe_clusters(
                clusters=['c2arn'], include=['STATISTICS']
            )
        ]
        c = cls.limits['Container Instances per Cluster'].get_current_usage()
        assert len(c) == 2
        assert c[0].get_value() == 11
        assert c[0].resource_id == 'c1name'
        assert c[1].get_value() == 3
        assert c[1].resource_id == 'c2name'
        s = cls.limits['Services per Cluster'].get_current_usage()
        assert len(s) == 2
        assert s[0].get_value() == 23
        assert s[0].resource_id == 'c1name'
        assert s[1].get_value() == 2
        assert s[1].resource_id == 'c2name'
        u = cls.limits['Clusters'].get_current_usage()
        assert len(u) == 1
        assert u[0].get_value() == 2
        assert u[0].resource_id is None
        f = cls.limits['Fargate Tasks'].get_current_usage()
        assert len(f) == 1
        assert f[0].get_value() == 4
        assert f[0].resource_id is None
        assert m_fuoc.mock_calls == [
            call(cls, 'c1name'),
            call(cls, 'c2name')
        ]

    def test_find_usage_one_cluster(self):

        def se_cluster(*_, **kwargs):
            if kwargs['services'] == ['s1arn']:
                return {
                    'services': [
                        {
                            'launchType': 'EC2',
                            'serviceName': 's1',
                            'desiredCount': 4
                        }
                    ]
                }
            elif kwargs['services'] == ['s2arn']:
                return {
                    'services': [
                        {
                            'launchType': 'Fargate',
                            'serviceName': 's2',
                            'desiredCount': 26
                        }
                    ]
                }
            elif kwargs['services'] == ['s3arn']:
                return {
                    'services': [
                        {
                            'launchType': 'EC2',
                            'serviceName': 's3',
                            'desiredCount': 8
                        }
                    ]
                }
            else:
                return {}

        mock_conn = Mock()
        mock_conn.describe_services.side_effect = se_cluster
        mock_paginator = Mock()
        mock_paginator.paginate.return_value = [{
            'serviceArns': [
                's1arn',
                's2arn',
                's3arn'
            ],
            'nextToken': 'string'
        }]
        mock_conn.get_paginator.return_value = mock_paginator

        cls = _EcsService(21, 43)
        cls.conn = mock_conn
        cls._find_usage_one_cluster('cName')

        assert mock_conn.mock_calls == [
            call.get_paginator('list_services'),
            call.get_paginator().paginate(
                cluster='cName', launchType='EC2'
            ),
            call.describe_services(cluster='cName', services=['s1arn']),
            call.describe_services(cluster='cName', services=['s2arn']),
            call.describe_services(cluster='cName', services=['s3arn'])
        ]
        u = cls.limits[
            'EC2 Tasks per Service (desired count)'
        ].get_current_usage()
        assert len(u) == 2
        assert u[0].get_value() == 4
        assert u[0].resource_id == 'cluster=cName; service=s1'
        assert u[0].aws_type == 'AWS::ECS::Service'
        assert u[1].get_value() == 8
        assert u[1].resource_id == 'cluster=cName; service=s3'
        assert u[1].aws_type == 'AWS::ECS::Service'

    def test_required_iam_permissions(self):
        cls = _EcsService(21, 43)
        assert sorted(cls.required_iam_permissions()) == [
            'ecs:DescribeClusters',
            'ecs:DescribeServices',
            'ecs:ListClusters',
            'ecs:ListServices'
        ]



"""
awslimitchecker/services/rds.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc  # noqa
import logging

from .base import _AwsService
from ..limit import AwsLimit

logger = logging.getLogger(__name__)


class _RDSService(_AwsService):

    service_name = 'RDS'
    api_name = 'rds'

    # Mapping of RDS DescribeAccountAttributes action AccountQuotaName string
    # to our Limit name
    API_NAME_TO_LIMIT = {
        'DBInstances': 'DB instances',
        'ReservedDBInstances': 'Reserved Instances',
        'AllocatedStorage': 'Storage quota (GB)',
        'DBSecurityGroups': 'DB security groups',
        'AuthorizationsPerDBSecurityGroup': 'Max auths per security group',
        'DBParameterGroups': 'DB parameter groups',
        'ManualSnapshots': 'DB snapshots per user',
        'EventSubscriptions': 'Event Subscriptions',
        'DBSubnetGroups': 'Subnet Groups',
        'OptionGroups': 'Option Groups',
        'SubnetsPerDBSubnetGroup': 'Subnets per Subnet Group',
        'ReadReplicasPerMaster': 'Read replicas per master',
        'DBClusters': 'DB Clusters',
        'DBClusterParameterGroups': 'DB Cluster Parameter Groups',
    }

    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update corresponding Limit via
        :py:meth:`~.AwsLimit._add_current_usage`.
        """
        logger.debug("Checking usage for service %s", self.service_name)
        self.connect()
        for lim in self.limits.values():
            lim._reset_usage()
        self._find_usage_instances()
        self._find_usage_subnet_groups()
        self._find_usage_security_groups()
        # RDS API also provides usage information
        self._update_limits_from_api()
        self._have_usage = True
        logger.debug("Done checking usage.")

    def _find_usage_instances(self):
        """find usage for DB Instances and related limits"""
        paginator = self.conn.get_paginator('describe_db_instances')
        for page in paginator.paginate():
            for instance in page['DBInstances']:
                self.limits['Read replicas per master']._add_current_usage(
                    len(instance['ReadReplicaDBInstanceIdentifiers']),
                    aws_type='AWS::RDS::DBInstance',
                    resource_id=instance['DBInstanceIdentifier']
                )

    def _find_usage_subnet_groups(self):
        """find usage for subnet groups"""
        paginator = self.conn.get_paginator('describe_db_subnet_groups')
        for page in paginator.paginate():
            for group in page['DBSubnetGroups']:
                self.limits['Subnets per Subnet Group']._add_current_usage(
                    len(group['Subnets']),
                    aws_type='AWS::RDS::DBSubnetGroup',
                    resource_id=group["DBSubnetGroupName"],
                )

    def _find_usage_security_groups(self):
        """find usage for security groups"""
        vpc_count = 0

        paginator = self.conn.get_paginator('describe_db_security_groups')
        for page in paginator.paginate():
            for group in page['DBSecurityGroups']:
                if 'VpcId' in group and group['VpcId'] is not None:
                    vpc_count += 1
                self.limits['Max auths per security group']._add_current_usage(
                    len(group["EC2SecurityGroups"]) + len(group["IPRanges"]),
                    aws_type='AWS::RDS::DBSecurityGroup',
                    resource_id=group['DBSecurityGroupName']
                )

        self.limits['VPC Security Groups']._add_current_usage(
            vpc_count,
            aws_type='AWS::RDS::DBSecurityGroup',
        )

    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        if self.limits != {}:
            return self.limits
        limits = {}
        limits['DB instances'] = AwsLimit(
            'DB instances',
            self,
            40,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBInstance'
        )
        limits['Reserved Instances'] = AwsLimit(
            'Reserved Instances',
            self,
            40,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBInstance',
        )
        limits['Storage quota (GB)'] = AwsLimit(
            'Storage quota (GB)',
            self,
            100000,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBInstance',
        )
        limits['DB snapshots per user'] = AwsLimit(
            'DB snapshots per user',
            self,
            100,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSnapshot',
        )
        limits['DB parameter groups'] = AwsLimit(
            'DB parameter groups',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBParameterGroup',
        )
        limits['DB security groups'] = AwsLimit(
            'DB security groups',
            self,
            25,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSecurityGroup',
        )
        limits['VPC Security Groups'] = AwsLimit(
            'VPC Security Groups',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSecurityGroup',
        )
        limits['Subnet Groups'] = AwsLimit(
            'Subnet Groups',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSubnetGroup',
            ta_limit_name='Subnet groups'
        )
        limits['Subnets per Subnet Group'] = AwsLimit(
            'Subnets per Subnet Group',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSubnetGroup',
            ta_limit_name='Subnets per subnet group'
        )
        limits['Option Groups'] = AwsLimit(
            'Option Groups',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBOptionGroup',
        )
        limits['Event Subscriptions'] = AwsLimit(
            'Event Subscriptions',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBEventSubscription',
            ta_limit_name='Event subscriptions'
        )
        limits['Read replicas per master'] = AwsLimit(
            'Read replicas per master',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBInstance',
        )
        # this is the number of rules per security group
        limits['Max auths per security group'] = AwsLimit(
            'Max auths per security group',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBSecurityGroup',
            limit_subtype='AWS::RDS::DBSecurityGroupIngress',
        )
        limits['DB Clusters'] = AwsLimit(
            'DB Clusters',
            self,
            40,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBCluster',
            ta_limit_name='Clusters'
        )
        limits['DB Cluster Parameter Groups'] = AwsLimit(
            'DB Cluster Parameter Groups',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::RDS::DBClusterParameterGroup',
            ta_limit_name='Cluster parameter groups'
        )
        self.limits = limits
        return limits

    def _update_limits_from_api(self):
        """
        Query RDS's DescribeAccountAttributes API action, and update limits
        with the quotas returned. Updates ``self.limits``.

        We ignore the usage information from the API,
        """
        self.connect()
        logger.info("Querying RDS DescribeAccountAttributes for limits")
        lims = self.conn.describe_account_attributes()['AccountQuotas']
        for lim in lims:
            if lim['AccountQuotaName'] not in self.API_NAME_TO_LIMIT:
                logger.info('RDS DescribeAccountAttributes returned unknown'
                            'limit: %s (max: %s; used: %s)',
                            lim['AccountQuotaName'], lim['Max'], lim['Used'])
                continue
            lname = self.API_NAME_TO_LIMIT[lim['AccountQuotaName']]
            self.limits[lname]._set_api_limit(lim['Max'])
            if len(self.limits[lname].get_current_usage()) < 1:
                self.limits[lname]._add_current_usage(lim['Used'])
        logger.debug('Done setting limits from API.')

    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        return [
            "rds:DescribeAccountAttributes",
            "rds:DescribeDBInstances",
            "rds:DescribeDBParameterGroups",
            "rds:DescribeDBSecurityGroups",
            "rds:DescribeDBSnapshots",
            "rds:DescribeDBSubnetGroups",
            "rds:DescribeEventSubscriptions",
            "rds:DescribeOptionGroups",
            "rds:DescribeReservedDBInstances",
        ]



"""
awslimitchecker/tests/services/test_s3.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from awslimitchecker.services.s3 import _S3Service
from awslimitchecker.limit import AwsLimit

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock
else:
    from unittest.mock import patch, call, Mock


pbm = 'awslimitchecker.services.s3'  # module patch base
pb = '%s._S3Service' % pbm  # class patch pase


class Test_S3Service(object):

    def test_init(self):
        """test __init__()"""
        cls = _S3Service(21, 43)
        assert cls.service_name == 'S3'
        assert cls.api_name == 's3'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _S3Service(21, 43)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Buckets',
        ])
        assert res['Buckets'].service == cls
        assert res['Buckets'].def_warning_threshold == 21
        assert res['Buckets'].def_critical_threshold == 43
        assert res['Buckets'].default_limit == 100

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock(spec_set=AwsLimit)
        cls = _S3Service(21, 43)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        mock_buckets = Mock()
        mock_buckets.all.return_value = ['a', 'b', 'c']
        mock_conn = Mock(buckets=mock_buckets)
        with patch('%s.connect_resource' % pb) as mock_connect:
            cls = _S3Service(21, 43)
            cls.resource_conn = mock_conn
            assert cls._have_usage is False
            cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is True
        assert mock_buckets.mock_calls == [call.all()]
        assert len(cls.limits['Buckets'].get_current_usage()) == 1
        assert cls.limits['Buckets'].get_current_usage()[0].get_value() == 3

    def test_required_iam_permissions(self):
        cls = _S3Service(21, 43)
        assert cls.required_iam_permissions() == [
            's3:ListAllMyBuckets'
        ]
